@@ -19,6 +19,9 @@ define([
      * Preliminary documentation for the REST API is at
      * https://github.com/ipython/ipython/wiki/IPEP-16%3A-Notebook-multi-directory-dashboard-and-URL-mapping#kernels-api
      * 
+     * Documentation for the messaging specifications is at
+     * https://jupyter-client.readthedocs.io/en/stable/messaging.html
+     *
      * @class Kernel
      * @param {string} kernel_service_url - the URL to access the kernel REST api
      * @param {string} ws_url - the websockets URL
@@ -30,6 +33,7 @@ define([
         this.id = null;
         this.name = name;
         this.ws = null;
+        this._stopping = false;
 
         this.kernel_service_url = kernel_service_url;
         this.kernel_url = null;
@@ -43,6 +47,7 @@ define([
         this.session_id = utils.uuid();
         this._msg_callbacks = {};
         this._msg_callbacks_overrides = {};
+        this._display_id_to_parent_ids = {};
         this._msg_queue = Promise.resolve();
         this.info_reply = {}; // kernel_info_reply stored here after starting
 
@@ -78,7 +83,7 @@ define([
                 username : this.username,
                 session : this.session_id,
                 msg_type : msg_type,
-                version : "5.0"
+                version : "5.2",
             },
             metadata : metadata || {},
             content : content,
@@ -131,11 +136,12 @@ define([
      * @function init_iopub_handlers
      */
     Kernel.prototype.init_iopub_handlers = function () {
-        var output_msg_types = ['stream', 'display_data', 'execute_result', 'error'];
+        var output_msg_types = ['stream', 'display_data', 'execute_result', 'error', 'update_display_data'];
         this._iopub_handlers = {};
         this.register_iopub_handler('status', $.proxy(this._handle_status_message, this));
         this.register_iopub_handler('clear_output', $.proxy(this._handle_clear_output, this));
         this.register_iopub_handler('execute_input', $.proxy(this._handle_input_message, this));
+        this.register_iopub_handler('shutdown_reply', $.proxy(this._handle_shutdown_message, this));
         
         for (var i=0; i < output_msg_types.length; i++) {
             this.register_iopub_handler(output_msg_types[i], $.proxy(this._handle_output_message, this));
@@ -149,7 +155,7 @@ define([
      *
      * @function list
      * @param {function} [success] - function executed on ajax success
-     * @param {function} [error] - functon executed on ajax error
+     * @param {function} [error] - function executed on ajax error
      */
     Kernel.prototype.list = function (success, error) {
         utils.ajax(this.kernel_service_url, {
@@ -175,7 +181,7 @@ define([
      * @function start
      * @param {params} [Object] - parameters to include in the query string
      * @param {function} [success] - function executed on ajax success
-     * @param {function} [error] - functon executed on ajax error
+     * @param {function} [error] - function executed on ajax error
      */
     Kernel.prototype.start = function (params, success, error) {
         var url = this.kernel_service_url;
@@ -215,7 +221,7 @@ define([
      *
      * @function get_info
      * @param {function} [success] - function executed on ajax success
-     * @param {function} [error] - functon executed on ajax error
+     * @param {function} [error] - function executed on ajax error
      */
     Kernel.prototype.get_info = function (success, error) {
         utils.ajax(this.kernel_url, {
@@ -233,13 +239,13 @@ define([
      *
      * Shutdown the kernel.
      *
-     * If you are also using sessions, then this function shoul NOT be
+     * If you are also using sessions, then this function should NOT be
      * used. Instead, use Session.delete. Otherwise, the session and
      * kernel WILL be out of sync.
      *
      * @function kill
      * @param {function} [success] - function executed on ajax success
-     * @param {function} [error] - functon executed on ajax error
+     * @param {function} [error] - function executed on ajax error
      */
     Kernel.prototype.kill = function (success, error) {
         this.events.trigger('kernel_killed.Kernel', {kernel: this});
@@ -261,7 +267,7 @@ define([
      *
      * @function interrupt
      * @param {function} [success] - function executed on ajax success
-     * @param {function} [error] - functon executed on ajax error
+     * @param {function} [error] - function executed on ajax error
      */
     Kernel.prototype.interrupt = function (success, error) {
         this.events.trigger('kernel_interrupting.Kernel', {kernel: this});
@@ -297,13 +303,14 @@ define([
          *
          * @function interrupt
          * @param {function} [success] - function executed on ajax success
-         * @param {function} [error] - functon executed on ajax error
+         * @param {function} [error] - function executed on ajax error
          */
         this.events.trigger('kernel_restarting.Kernel', {kernel: this});
         this.stop_channels();
-
         this._msg_callbacks = {};
         this._msg_callbacks_overrides = {};
+        this._display_id_to_parent_ids = {};
+
         var that = this;
         var on_success = function (data, status, xhr) {
             that.events.trigger('kernel_created.Kernel', {kernel: that});
@@ -314,7 +321,7 @@ define([
         };
 
         var on_error = function (xhr, status, err) {
-            that.events.trigger('kernel_dead.Kernel', {kernel: that});
+            that.events.trigger('kernel_failed_restart.Kernel', {kernel: that});
             that._kernel_dead();
             if (error) {
                 error(xhr, status, err);
@@ -454,7 +461,7 @@ define([
         var ws_host_url = this.ws_url + this.kernel_url;
 
         console.log("Starting WebSockets:", ws_host_url);
-        
+
         this.ws = new this.WebSocket([
                 that.ws_url,
                 utils.url_path_join(that.kernel_url, 'channels'),
@@ -464,33 +471,31 @@ define([
         
         var already_called_onclose = false; // only alert once
         var ws_closed_early = function(evt){
+            console.log("WebSocket closed early", evt);
             if (already_called_onclose){
                 return;
             }
             already_called_onclose = true;
-            if ( ! evt.wasClean ){
-                // If the websocket was closed early, that could mean
-                // that the kernel is actually dead. Try getting
-                // information about the kernel from the API call --
-                // if that fails, then assume the kernel is dead,
-                // otherwise just follow the typical websocket closed
-                // protocol.
-                that.get_info(function () {
-                    that._ws_closed(ws_host_url, false);
-                }, function () {
-                    that.events.trigger('kernel_dead.Kernel', {kernel: that});
-                    that._kernel_dead();
-                });
-            }
+            // If the websocket was closed early, that could mean
+            // that the kernel is actually dead. Try getting
+            // information about the kernel from the API call --
+            // if that fails, then assume the kernel is dead,
+            // otherwise just follow the typical websocket closed
+            // protocol.
+            that.get_info(function () {
+                that._ws_closed(ws_host_url, false);
+            }, function () {
+                that.events.trigger('kernel_dead.Kernel', {kernel: that});
+                that._kernel_dead();
+            });
         };
         var ws_closed_late = function(evt){
+            console.log("WebSocket closed unexpectedly", evt);
             if (already_called_onclose){
                 return;
             }
             already_called_onclose = true;
-            if ( ! evt.wasClean ){
-                that._ws_closed(ws_host_url, false);
-            }
+            that._ws_closed(ws_host_url, false);
         };
         var ws_error = function(evt){
             if (already_called_onclose){
@@ -505,7 +510,7 @@ define([
         this.ws.onerror = ws_error;
         // switch from early-close to late-close message after 1s
         setTimeout(function() {
-            if (that.ws !== null) {
+            if (that.ws !== null && !that._stopping) {
                 that.ws.onclose = ws_closed_late;
             }
         }, 1000);
@@ -576,11 +581,14 @@ define([
          */
         var that = this;
         var close = function () {
+            that._stopping = false;
             if (that.ws && that.ws.readyState === WebSocket.CLOSED) {
                 that.ws = null;
             }
         };
         if (this.ws !== null) {
+            // flag to avoid races with on_close_late
+            this._stopping = true;
             if (this.ws.readyState === WebSocket.OPEN) {
                 this.ws.onclose = close;
                 this.ws.close();
@@ -635,7 +643,7 @@ define([
         } else {
             this._pending_messages.push(msg);
         }
-    }
+    };
     
     Kernel.prototype.send_shell_message = function (msg_type, content, callbacks, metadata, buffers) {
         /**
@@ -730,6 +738,7 @@ define([
          *      @param callbacks.iopub.output {function}
          *      @param callbacks.iopub.clear_output {function}
          *      @param callbacks.input {function}
+         *      @param callbacks.clear_on_done=true {Boolean}
          * @param {object} [options]
          *      @param [options.silent=false] {Boolean}
          *      @param [options.user_expressions=empty_dict] {Dict}
@@ -764,7 +773,7 @@ define([
          *      }
          *
          * Each callback will be passed the entire message as a single
-         * arugment.  Payload handlers will be passed the corresponding
+         * argument.  Payload handlers will be passed the corresponding
          * payload and the execute_reply message.
          */
         var content = {
@@ -886,6 +895,25 @@ define([
      */
     Kernel.prototype.clear_callbacks_for_msg = function (msg_id) {
         if (this._msg_callbacks[msg_id] !== undefined ) {
+            var callbacks = this._msg_callbacks[msg_id];
+            var kernel = this;
+            // clear display_id:msg_id map for display_ids associated with this msg_id
+            if (!callbacks) return;
+            callbacks.display_ids.map(function (display_id) {
+                var msg_ids = kernel._display_id_to_parent_ids[display_id];
+                if (msg_ids) {
+                    var idx = msg_ids.indexOf(msg_id);
+                    if (idx === -1) {
+                        return;
+                    }
+                    if (msg_ids.length === 1) {
+                        delete kernel._display_id_to_parent_ids[display_id];
+                    } else {
+                        msg_ids.splice(idx, 1);
+                        kernel._display_id_to_parent_ids[display_id] = msg_ids;
+                    }
+                }
+            });
             delete this._msg_callbacks[msg_id];
         }
     };
@@ -897,7 +925,7 @@ define([
         var callbacks = this._msg_callbacks[msg_id];
         if (callbacks !== undefined) {
             callbacks.shell_done = true;
-            if (callbacks.iopub_done) {
+            if (callbacks.clear_on_done && callbacks.iopub_done) {
                 this.clear_callbacks_for_msg(msg_id);
             }
         }
@@ -910,10 +938,11 @@ define([
         var callbacks = this._msg_callbacks[msg_id];
         if (callbacks !== undefined) {
             callbacks.iopub_done = true;
-            if (callbacks.shell_done) {
+            if (callbacks.clear_on_done && callbacks.shell_done) {
                 this.clear_callbacks_for_msg(msg_id);
             }
         }
+        this.events.trigger('finished_iopub.Kernel', {kernel: this, msg_id: msg_id});
     };
     
     /**
@@ -939,8 +968,14 @@ define([
             cbcopy.shell = callbacks.shell;
             cbcopy.iopub = callbacks.iopub;
             cbcopy.input = callbacks.input;
+            cbcopy.clear_on_done = callbacks.clear_on_done;
             cbcopy.shell_done = (!callbacks.shell);
             cbcopy.iopub_done = (!callbacks.iopub);
+            cbcopy.display_ids = [];
+            if (callbacks.clear_on_done === undefined) {
+                // default to clear-on-done
+                cbcopy.clear_on_done = true;
+            }
         } else if (remember) {
             this.last_msg_callbacks = {};
         }
@@ -951,7 +986,7 @@ define([
      */
     Kernel.prototype.output_callback_overrides_push = function(msg_id, callback_id) {
         var output_callbacks = this._msg_callbacks_overrides[msg_id];
-        if (output_callbacks === void 0) {
+        if (!output_callbacks) {
             this._msg_callbacks_overrides[msg_id] = output_callbacks = [];
         }
         output_callbacks.push(callback_id);
@@ -1106,8 +1141,52 @@ define([
      * @function _handle_output_message
      */
     Kernel.prototype._handle_output_message = function (msg) {
+        var that = this;
         var msg_id = msg.parent_header.msg_id;
         var callbacks = this.get_output_callbacks_for_msg(msg_id);
+        if (['display_data', 'update_display_data', 'execute_result'].indexOf(msg.header.msg_type) > -1) {
+            // display_data messages may re-route based on their display_id
+            var display_id = (msg.content.transient || {}).display_id;
+            if (display_id) {
+                // it has a display_id
+                var parent_ids = this._display_id_to_parent_ids[display_id];
+                if (parent_ids) {
+                    // we've seen it before, update existing outputs with same display_id
+                    // by handling display_data as update_display_data
+                    var update_msg = $.extend(true, {}, msg);
+                    update_msg.header.msg_type = 'update_display_data';
+
+                    parent_ids.map(function (parent_id) {
+                        var callbacks = that.get_callbacks_for_msg(parent_id);
+                        if (!callbacks) return;
+                        var callback = callbacks.iopub.output;
+                        if (callback) {
+                            callback(update_msg);
+                        }
+                    });
+                }
+                // we're done here if it's update_display
+                if (msg.header.msg_type === 'update_display_data') {
+                    // it's an update, don't proceed to the normal display
+                    return;
+                }
+                // regular display_data with id, record it for future updating
+                // in _display_id_to_parent_ids for future lookup
+                if (this._display_id_to_parent_ids[display_id] === undefined) {
+                    this._display_id_to_parent_ids[display_id] = [];
+                }
+                var callback_id = this.get_output_callback_id(msg_id);
+                if (this._display_id_to_parent_ids[display_id].indexOf(callback_id) === -1) {
+                    this._display_id_to_parent_ids[display_id].push(callback_id);
+                }
+                // and in callbacks for cleanup on clear_callbacks_for_msg
+                if (callbacks && callbacks.display_ids.indexOf(display_id) === -1) {
+                    callbacks.display_ids.push(display_id);
+                }
+            }
+        }
+
+
         if (!callbacks || !callbacks.iopub) {
             // The message came from another client. Let the UI decide what to
             // do with it.
@@ -1134,6 +1213,19 @@ define([
         }
     };
 
+
+    /**
+     * Handle a kernel shutdown message
+     * @function _handle_shutdown_message
+     */
+    Kernel.prototype._handle_shutdown_message = function (msg) {
+        if (!msg.content.restart) {
+            this.events.trigger('kernel_dead.Kernel', {kernel: this});
+            this._kernel_dead();
+        }
+    }
+
+
     /**
      * Dispatch IOPub messages to respective handlers. Each message
      * type should have a handler.
